/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

/**
 * @file      command.cpp
 * @author    Bruno ANSELME
 * @brief     Command class definition
 * @details   Command types, mouse types and meta types definitions.\n
 *            Types translation to text and text to type.
 */

#include <QRegularExpression>

#include "app_version.h"
#include "command.h"

static QMap<Command::comType,   QString>    type_name;          // Labels for Command::comType
static QMap<Command::metaType,  QString>    meta_name;          // Labels for Command::metaType
static QMap<Command::metaType,  QString>    meta_desc;          // Descriptions for Command::metaType
static QMap<Command::mouseType, QString>    mouse_name;         // Labels for Command::mouseType
static QMap<QString,    Command::comType>   type_key;           // Reversed type_name
static QMap<QString,    Command::metaType>  meta_key;           // Reversed meta_name
static QMap<QString,    Command::mouseType> mouse_key;          // Reversed mouse_name

void Command::initCommands() {
    if (type_name.isEmpty()) {
        type_name[Meta]            = QString("Meta");
        type_name[Key]             = QString("Key");
        type_name[Mouse]           = QString("Mouse");
        type_name[Text]            = QString("Text");
        type_name[Exec]            = QString("Exec");
        QMapIterator<comType, QString> i(type_name);
        while (i.hasNext()) { i.next(); type_key[i.value()] = i.key(); }
    }
    if (meta_name.isEmpty()) {
        meta_name[Quit]            = QString("Quit %1").arg(APP_NAME);
        meta_name[Stop]            = QString("Stop");
        meta_name[Suspend]         = QString("Suspend");
        meta_name[Resume]          = QString("Resume");
        meta_name[Repeat]          = QString("Repeat");
        meta_name[Settings]        = QString("Settings");
        meta_name[Reminder]        = QString("Reminder");
        meta_name[Raise]           = QString("Raise");
        meta_name[Minimize]        = QString("Minimize");
        meta_name[Maximize]        = QString("Maximize");
        meta_name[MouseShow]       = QString("MouseShow");
        meta_name[StartDictation]  = QString("StartDictation");
        meta_name[StopDictation]   = QString("StopDictation");
        meta_name[OneDictation]    = QString("OneDictation");
        meta_name[SetsSelection]   = QString("SetsSelection");
        meta_name[Speak]           = QString("Speak");
        QMapIterator<metaType, QString> i(meta_name);
        while (i.hasNext()) { i.next(); meta_key[i.value()] = i.key(); }
    }
    if (meta_desc.isEmpty()) {
        meta_desc[Quit]            = QObject::tr("Quit %1").arg(APP_NAME);
        meta_desc[Stop]            = QObject::tr("Stop listening");
        meta_desc[Suspend]         = QObject::tr("Suspend listening");
        meta_desc[Resume]          = QObject::tr("Resume listening");
        meta_desc[Repeat]          = QObject::tr("Repeat last valid command");
        meta_desc[Settings]        = QObject::tr("Open settings dialog");
        meta_desc[Reminder]        = QObject::tr("Show full screen reminder");
        meta_desc[Raise]           = QObject::tr("Raise %1 window").arg(APP_NAME);
        meta_desc[Minimize]        = QObject::tr("Minimize current window");
        meta_desc[Maximize]        = QObject::tr("Maximize current window");
        meta_desc[MouseShow]       = QObject::tr("Show mouse position");
        meta_desc[StartDictation]  = QObject::tr("Enter dictation mode");
        meta_desc[StopDictation]   = QObject::tr("Exit dictation mode");
        meta_desc[OneDictation]    = QObject::tr("Dictate only once");
        meta_desc[SetsSelection]   = QObject::tr("Open sets selection tab");
        meta_desc[Speak]           = QObject::tr("Text to speech, vocalize clipboard contents");
        QMapIterator<metaType, QString> i(meta_desc);
        while (i.hasNext()) { i.next(); meta_key[i.value()] = i.key(); }
    }
    if (mouse_name.isEmpty()) {
        mouse_name[Left]           = QString("Click left");
        mouse_name[Middle]         = QString("Click middle");
        mouse_name[Right]          = QString("Click right");
        mouse_name[DownLeft]       = QString("Press left");
        mouse_name[DownMiddle]     = QString("Press middle");
        mouse_name[DownRight]      = QString("Press right");
        mouse_name[UpLeft]         = QString("Release left");
        mouse_name[UpMiddle]       = QString("Release middle");
        mouse_name[UpRight]        = QString("Release right");
        mouse_name[WheelUp]        = QString("Wheel up");
        mouse_name[WheelDown]      = QString("Wheel down");
        mouse_name[MoveUp]         = QString("Move up");
        mouse_name[MoveDown]       = QString("Move down");
        mouse_name[MoveLeft]       = QString("Move left");
        mouse_name[MoveRight]      = QString("Move right");
        mouse_name[MoveTo]         = QString("Move to");
        mouse_name[GlideUp]        = QString("Glide up");
        mouse_name[GlideDown]      = QString("Glide down");
        mouse_name[GlideLeft]      = QString("Glide left");
        mouse_name[GlideRight]     = QString("Glide right");
        mouse_name[GlideUpLeft]    = QString("Glide up left");
        mouse_name[GlideUpRight]   = QString("Glide up right");
        mouse_name[GlideDownLeft]  = QString("Glide down left");
        mouse_name[GlideDownRight] = QString("Glide down right");
        QMapIterator<mouseType, QString> i(mouse_name);
        while (i.hasNext()) { i.next(); mouse_key[i.value()] = i.key(); }
    }
}

//----------------------------
Command::Command(comType typ) :   cmdtype(typ) { }

Command *Command::createKeyCommand(const QString &keyseq) {
    Command *cmd    = new Command(Key);
    cmd->setQtKey(keyseq);
    cmd->qt2keysym();
    return cmd;
}

Command *Command::createMetaCommand(metaType mc){
    Command *cmd    = new Command(Meta);
    cmd->setMetaType(mc);
    cmd->setQtKey(QString(meta_name[cmd->getMetaType()]));
    cmd->setKeySym(cmd->getQtKey());
    return cmd;
}

Command *Command::createMouseCommand(mouseType ms){
    Command *cmd    = new Command(Mouse);
    cmd->setMouseType(ms);
    cmd->setQtKey(QString(mouse_name[cmd->getMouseType()]));
    cmd->setKeySym(cmd->getQtKey());
    return cmd;
}

Command *Command::createTextCommand(const QString &textToSend) {
    Command *cmd = new Command(Text);
    cmd->setQtKey(textToSend);
    cmd->setKeySym(cmd->getQtKey());
    return cmd;
}

Command *Command::createExecCommand(const QString &application) {
    Command *cmd = new Command(Exec);
    cmd->setQtKey(application);
    cmd->setKeySym(cmd->getQtKey());
    return cmd;
}

QString Command::addQuotes(const QString s) {
    QRegularExpression re("^\".*\"$");
    QRegularExpressionMatch match = re.match(s);
    if (match.hasMatch()) {
        return s;
    } else {
        return QString("\"%1\"").arg(s);
    }
}

QString Command::removeQuotes(const QString s) {
    QRegularExpression re("^\"(.*)\"$");
    QRegularExpressionMatch match = re.match(s);
    if (match.hasMatch()) {
        return match.captured(1);
    }
    return s;
}

const QString &Command::getTypeName()               { return type_name[cmdtype];    }
const QString &Command::getMetaName()               { return meta_name[meta];       }
const QString &Command::getMetaDescription()        { return meta_desc[meta];       }
const QString &Command::getMouseName()              { return mouse_name[mouse];     }

const QString &Command::getTypeName(comType t)      { return type_name[t];          }
const QString &Command::getMetaName(metaType m)     { return meta_name[m];          }
const QString &Command::getMouseName(mouseType m)   { return mouse_name[m];         }

QStringList Command::getTypesNames() {
    QStringList lst;
    QMapIterator<comType, QString> i(type_name);
    while (i.hasNext()) { i.next(); lst.append(i.value()); }
    return lst;
}

QStringList Command::getMetaNames() {
    QStringList lst;
    QMapIterator<metaType, QString> i(meta_name);
    while (i.hasNext()) { i.next(); lst.append(i.value()); }
    return lst;
}

QStringList Command::getMetaDescriptions() {
    QStringList lst;
    QMapIterator<metaType, QString> i(meta_desc);
    while (i.hasNext()) { i.next(); lst.append(i.value()); }
    return lst;
}

QStringList Command::getMouseNames() {
    QStringList lst;
    QMapIterator<mouseType, QString> i(mouse_name);
    while (i.hasNext()) { i.next(); lst.append(i.value()); }
    return lst;
}

Command::comType     Command::typeFromName(QString name)    { return type_key[name];   }
Command::metaType    Command::metaFromName(QString name)    { return meta_key[name];   }
Command::mouseType   Command::mouseFromName(QString name)   { return mouse_key[name];  }

void Command::qt2keysym() {     // Convert QKeySequence to KeySyms sequence
    if (cmdtype == Key) {
        cmdkeysym   = cmdqt;
//        cmdkeysym   = guiKeySequence(cmdqt);
    }
}
