/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

/**
 * @file      comprendo.cpp
 * @author    Bruno ANSELME
 * @brief     Comprendo class is the application's engine
 * @details   Load several commands sets together (selection from settings).\n
 *              Call voice recognition.\n
 *              Find command from utterance.\n
 *              Manage mouse glide functions.\n
 *              Manage mouse zoom functions.\n
 *              Performs Key, Text, Mouse and Exec commands types (Comprendo::doCommands()).\n
 *              Meta-commands send signals (will be connected to MainWindow::execMeta).\n
 */

#include <QCommandLineParser>
#include <QRegularExpression>
#include <QJsonObject>
#include <QJsonDocument>

#include "app_version.h"
#include "misc.h"
#include "comprendo.h"
#include "zoombutton.h"

//----------------------------
Comprendo::Comprendo(QObject *parent) : QObject(parent)
  , cmdSets()
  , lastCommand()
  , lastCount(1)
  , doTool(nullptr)
  , listening(false)
  , status(VoskStatus::Empty)
  , format()
  , audio()
  , audioInput()
{
    // Set up audio and vosk
    microphone = nullptr;
    format.setSampleRate(44100);
    format.setChannelCount(1);
    format.setSampleFormat(QAudioFormat::Int16);

    QAudioDevice info = QMediaDevices::defaultAudioInput();
    if (!info.isFormatSupported(format)) {
        qWarning() << "Default format not supported, trying to use the nearest.";
    }
    audio = new QAudioSource(audioInput.device(), format, this);

    restartDotool();

    vosk_set_log_level(-1);            // Turn off vosk_api output (-1 ... 4 or more)
    model = vosk_model_new(APPCONF.model.toStdString().c_str());
    recognizer = vosk_recognizer_new(model, audio->format().sampleRate());
    // vosk_recognizer_set_endpointer_delays(recognizer, 5.0, 0.5, 20);

    timerGlide  = 0;
    zooming     = false;
    direction   = None;
    screen = getDesktopRect();
    connect(this,           SIGNAL(stopGlide()),       this,   SLOT(stopGliding()));
    loadAllSets();
    emit statusListening(false);
}

Comprendo::~Comprendo() {
    if (listening) {
        audio->stop();
        microphone = nullptr;
    }
    vosk_recognizer_free(recognizer);
    vosk_model_free(model);
    delete doTool;
    delete audio;
}

void Comprendo::setAudioDevice(QByteArray id) {
    const QList<QAudioDevice> audioDevices = QMediaDevices::audioInputs();
    bool memListen = listening;
    listen(false);
    for (const QAudioDevice &device : audioDevices) {
        if (id == device.id()) {
            APPCONF.audioId = id;
            audioInput.setDevice(device);
            delete audio;
            audio = new QAudioSource(audioInput.device(), format, this);
            // audio->start();
            break;
        }
    }
    listen(memListen);
}

void Comprendo::restartDotool() {
    if (doTool)
        delete doTool;
    doTool = new DoToolPipe(APPCONF.keyboardVariant);
}

void Comprendo::listen(bool b) {
    listening = b;
    if (listening) {
        if (!microphone)
            microphone = audio->start();
    } else {
        vosk_recognizer_result(recognizer);     // flush recognizer
        vosk_recognizer_reset(recognizer);
        audio->stop();
        microphone = nullptr;
    }
}

QStringList Comprendo::getSetFiles(QString lang) {
    QSettings settings(APP_ORG, APP_NAME);      // Read sets settings here
    settings.beginGroup(lang);
    QStringList keys = settings.allKeys();
    settings.endGroup();
    return keys;
}

void Comprendo::loadAllSets() {
    cmdSets.clear();
    QStringList  cmds = Comprendo::getSetFiles(APPCONF.setsDirName);
    QListIterator<QString> cmd(cmds);
    while (cmd.hasNext()) {
        cmdSets.append(new CommandSet(QString("%1/%2.noc").arg(APPCONF.userSetsPath.absolutePath()).arg(cmd.next())));
    }

    QListIterator<CommandSet *> comds(cmdSets);
    while (comds.hasNext()) {
        comds.next()->loadCommands();
    }
}

QString Comprendo::getCommandsList() {          // Return commands utterances with sets name for reminder
    QStringList ret;
    ret.append(QString("-- %1").arg(tr("Repeated utterance")));
    ret.append(lastCommand);
    QListIterator<CommandSet *> cmds(cmdSets);
    while (cmds.hasNext()) {
        CommandSet *set = cmds.next();
        ret.append(QString("-- %1").arg(set->getName()));
        ret.append(set->getCommandsUtterance());
    }
    return ret.join("\n");
}

QString Comprendo::getCommandsUtterance() {     // Return commands utterances for .lm
    QStringList ret;
    QListIterator<CommandSet *> cmds(cmdSets);
    while (cmds.hasNext()) {
        ret.append(cmds.next()->getCommandsUtterance());
    }
    ret.sort();
    return ret.join("\n");
}

QString Comprendo::getCommandsCSV() {           // Return commands for CSV
    QStringList ret;
    QListIterator<CommandSet *> cmds(cmdSets);
    while (cmds.hasNext()) {
        ret.append(cmds.next()->getCommandsCSV());
    }
    ret.sort();
    return ret.join("\n");
}

void Comprendo::fillTable(QTableWidget *table, CommandSet::ShortcutFormat format) {     // Fill GUI table
    QListIterator<CommandSet *> cmds(cmdSets);
    while (cmds.hasNext()) {
        cmds.next()->fillTable(table, format);
    }
}

void Comprendo::repeatLastCommand(bool suspended) {
    int count(1);
    Command *com = getCommand(lastCommand, count);
    if (com)
        execCommand(lastCommand, com, lastCount, suspended);
}

void Comprendo::stopGliding() {
    killTimer(timerGlide);
    timerGlide  = 0;
    direction   = None;
}

void Comprendo::timerEvent(QTimerEvent *t){
    if (t->timerId()==timerGlide) {     // Timer for mouse glide
        switch (direction) {
        case Left:      doTool->mouseAction(Command::GlideLeft);     break;
        case Right:     doTool->mouseAction(Command::GlideRight);    break;
        case Up:        doTool->mouseAction(Command::GlideUp);       break;
        case Down:      doTool->mouseAction(Command::GlideDown);     break;
        case UpLeft:    doTool->mouseAction(Command::GlideUpLeft);   break;
        case UpRight:   doTool->mouseAction(Command::GlideUpRight);  break;
        case DownLeft:  doTool->mouseAction(Command::GlideDownLeft); break;
        case DownRight: doTool->mouseAction(Command::GlideDownRight);break;
        case None:
        default: break;
        }

//        int x, y, screenNum;
        QRect screen = getDesktopRect();
        int x = QCursor::pos().x();
        int y = QCursor::pos().y();

//        guiGetMousePos(&x, &y, &screenNum);
        if ((x <= screen.left()) || (x >= screen.right()) || (y <= screen.top()) || (y >= screen.bottom()))
            stopGliding();
    }
}

void Comprendo::setMouseSteps(int moveStep_, int glideStep_, int wheelStep_) {
    doTool->setSteps(moveStep_, glideStep_, wheelStep_);
}

void Comprendo::setKeyDelays(int keyDelay, int keyHold, int typeDelay, int typeHold) {
    doTool->setDelays(keyDelay, keyHold, typeDelay, typeHold);
}

void Comprendo::writeText(QString str) {
    doTool->writePipe(QString("type %1").arg(str));
}

void Comprendo::writeRaw(QString str) {
    doTool->writePipe(str);
}

void Comprendo::writeLongText(QString str) {
    QStringList pieces = str.split(QRegularExpression("(\\\\n|\\n)"));  // Either \n or \\n
    for (int k = 0; k < pieces.count(); k++) {
        QStringList smallPieces = pieces[k].split("\\t");
        for (int kk = 0; kk < smallPieces.count(); kk++) {
            doTool->writePipe(QString("type %1").arg(smallPieces[kk]));
            if (kk < smallPieces.count() - 1)
                doTool->writePipe("key tab");
        }
        if (k < pieces.count() - 1)
            doTool->writePipe("key enter");
    }
}

Command *Comprendo::getCommand(QString cmd, int &count) {
    count = 1;
    QString command(cmd);
    // Try to find multiplier
    for (int i = 0; i < repeaters.count(); i++) {
        QRegularExpression mult(QString("(.*) %1$").arg(repeaters[i]));

        QRegularExpressionMatch match = mult.match(cmd);
        if (match.hasMatch()) {
            command = match.captured(1);            // command utterance
            count = i + 1;
            break;
        }
    }
    lastCount = count;

    // Process command
    QListIterator<CommandSet *> cmds(cmdSets);
    while (cmds.hasNext()) {
        const QMap<QString, Command *> *commands = cmds.next()->getCommands();
        if (commands->contains(command)) {
            return commands->value(command);
        }
    }
    return nullptr;
}

#define MOUSEGLIDE(WHERE) direction = WHERE; timerGlide  = startTimer(APPCONF.glideRate); line.append(" : ").append(com->getMouseName());

QString Comprendo::execCommand(QString &cmd, Command *com, int count, bool suspended) {
    QString         memo = cmd;
    QString         line = cmd.leftJustified(20,' ');
    switch (com->getType()) {
    case Command::Meta :            // Meta commands
    {   // No count with meta commands (except Repeat).
        Command::metaType typ = com->getMetaType();
        line.append(" : ").append(com->getMetaName());
        if ((suspended) && (typ != Command::Resume)) break;
        if (typ == Command::Repeat)
            line.append(" \"").append(lastCommand).append("\"");
    }
        break;
    case Command::Key :             // Keyboard shortcuts, and shorcuts sequences
        if (suspended) {
            line.clear();
        } else {
            lastCommand = memo;
            line.append(" : ").append(SEQ_TO_NATIVE(com->getQtKey()).toStdString().c_str());
            if (count>1) line.append(QString("(%1)").arg(count));
            for (int j = 0; j < count; j++) {
                QStringList coms = com->getKeySym().split(", ");
                for (int n = 0; n < coms.count(); n++) {
                    doTool->sendSequence(coms[n]);
                }
            }
        }
        break;
    case Command::Mouse :           // Mouse commands
        if (suspended) {
            line.clear();
        } else {
            lastCommand = memo;
            Command::mouseType mous = com->getMouseType();
            switch (mous) {
            //  Mouse click
            case Command::Left:
            case Command::Middle:
            case Command::Right:
                //  Mouse wheel
            case Command::WheelUp:
            case Command::WheelDown:
                // Mouse relatives moves
            case Command::MoveLeft:
            case Command::MoveRight:
            case Command::MoveUp:
            case Command::MoveDown:
                //  Mouse down commands
            case Command::DownLeft:
            case Command::DownMiddle:
            case Command::DownRight:
                //  Mouse up commands
            case Command::UpLeft:
            case Command::UpMiddle:
            case Command::UpRight:
                doTool->mouseAction(mous, count);
                line.append(" : ").append(com->getMouseName()); if (count>1) line.append(QString("(%1)").arg(count));
                break;
            case Command::MoveTo:
            {
                line.append(" : ").append(com->getMouseName());
                zooming = true;
                ZoomButton w(this);
                w.exec();
                w.hide();
                if (w.hasClicked()) {
                    doTool->mouseTo(w.whereRatio().x(), w.whereRatio().y());
                    line.append(QString(" %1,%2").arg(w.where().x()).arg(w.where().y()));
                }
                zooming = false;
            }
                break;
            // Mouse glide commands
            case Command::GlideLeft:        MOUSEGLIDE(Left)        break;
            case Command::GlideRight:       MOUSEGLIDE(Right)       break;
            case Command::GlideUp:          MOUSEGLIDE(Up)          break;
            case Command::GlideDown:        MOUSEGLIDE(Down)        break;
            case Command::GlideUpLeft:      MOUSEGLIDE(UpLeft)      break;
            case Command::GlideUpRight:     MOUSEGLIDE(UpRight)     break;
            case Command::GlideDownLeft:    MOUSEGLIDE(DownLeft)    break;
            case Command::GlideDownRight:   MOUSEGLIDE(DownRight)   break;
            default: break;
            }
        }
        break;
    case Command::Text :            // Text input command
        if (suspended) {
            line.clear();
        } else {
            lastCommand = memo;
            QString txt(com->getText());
            line.append(" : \"").append(txt.toStdString().c_str()).append("\"");
            if (count>1) line.append(QString("(%1)").arg(count));
            if (!txt.isEmpty()) {
                txt.replace(QString("\\\""), QString("\""));
                for (int j = 0; j < count; j++) {
                    writeLongText(txt);
                }
            }
        }
        break;
    case Command::Exec :            // Exec command
        if (suspended) {
            line.clear();
        } else {
            lastCommand = memo;
            line.append(" : ").append(com->getExec().toStdString().c_str());
            if (count > 1) line.append(QString("(%1)").arg(count));
            for (int j = 0; j < count; j++) {
                int pos = com->getExec().indexOf(" ");
                const QString exe  = com->getExec().left(pos);
                QString args;
                if (pos > 0)
                    args    = com->getExec().right(com->getExec().length() - pos);
                const QStringList argsList = parseCombinedArgString(args);
                QProcess *newProc = new QProcess();
                newProc->start(exe, argsList);
            }
        }
        break;
    }
    return line;
}

#undef MOUSEGLIDE

QString *Comprendo::recognize_from_microphone() {
    QString *returnStr = new QString();               // Pointer must be deleted by the caller
    status = VoskStatus::Empty;
    if (microphone) {
        qsizetype length =  audio->bytesAvailable();
        QByteArray buf = microphone->readAll();
        QJsonObject jsonObj;
        QRegularExpression re("(\\d),(\\d)");
        int res = vosk_recognizer_accept_waveform(recognizer, buf.constData(), length);
        if (res) {  // Final utterance
            QString str = vosk_recognizer_result(recognizer);
            jsonObj = QJsonDocument::fromJson(str.toUtf8()).object();
            returnStr->append(jsonObj["text"].toString());
            emit statusListening(true);
            if (returnStr->isEmpty()) {
                status = VoskStatus::Empty;
            } else {
                if (timerGlide)
                    emit stopGlide();
                status = VoskStatus::Final;
            }
        } else  {   // Partial utterance
            QString str =vosk_recognizer_partial_result(recognizer);
            jsonObj = QJsonDocument::fromJson(str.toUtf8()).object();
            returnStr->append(jsonObj["partial"].toString());
            if (returnStr->isEmpty()) {
                status = VoskStatus::Empty;
                emit statusListening(false);
            } else {
                if (timerGlide)
                    emit stopGlide();
                status = VoskStatus::Partial;
                emit statusListening(true);
            }
        }
    }
    return returnStr;
}
