/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

#ifndef COMPRENDO_H
#define COMPRENDO_H

#include <QObject>
#include <QString>
#include <QMap>
#include <QDir>
#include <QProcess>
#include <QAudioInput>
#include <QAudioSource>
#include <QMediaDevices>

#include "commandset.h"
#include "config.h"
#include "vosk_api.h"
#include "dotoolpipe.h"

class Comprendo : public QObject {
    Q_OBJECT
public:
    enum        ShortcutFormat { QtShortcut, XdoShortcut };
    enum        Direction { None, Left, Right, Up, Down, UpLeft, UpRight, DownLeft, DownRight };
    enum        VoskStatus { Empty, Partial, Final };
    explicit    Comprendo(QObject *parent = nullptr);                  //!< Create object from parameters
    ~Comprendo();
    static  QStringList     getSetFiles(QString lang);  //!< Get list of selected commands sets

    Command    *getCommand(QString cmd, int &count);                                //!< Return Command object from utterance
    QString     execCommand(QString &cmd, Command *com, int count, bool suspended); //!< Run Command. Returns a line to display
    QString     getCommandsList();                                                  //!< Return formatted commands list for Reminder
    QString     getCommandsUtterance();                                             //!< Return commands list for language model
    QString     getCommandsCSV();                                                   //!< Return formatted commands list for CSV
    QString     getLastCommand()            { return lastCommand; }                 //!< Return last command string
    void        repeatLastCommand(bool suspended);                                  //!< Repeat last command and return last command string
    void        fillTable(QTableWidget *table, CommandSet::ShortcutFormat format);  //!< Fill GUI table with commands names and actions. Calls CommandSet::fillTable
    bool        isZooming()                 { return zooming; }                     //!< True when mouse zooming in progress
    bool        isGliding()                 { return (direction != None); }         //!< True when mouse gliding in progress
    VoskStatus  getStatus()                 { return status; }
    bool        isListening()               { return listening; }
    void        resetVosk()                 { vosk_recognizer_reset(recognizer); }
    void        setMouseSteps(int moveStep_, int glideStep_, int wheelStep_);
    void        setKeyDelays(int keyDelay, int keyHold, int typeDelay, int typeHold);
    void        writeText(QString);
    void        writeRaw(QString);
    void        writeLongText(QString);
    void        setRepeaters(QStringList repeaters_)    { repeaters = repeaters_; }
    void        restartDotool();
    void        setAudioDevice(QByteArray id);

Q_SIGNALS:
    void        stopGlide();                                                        //!< Mouse glide asked to stopped
    void        statusListening(bool);                                              //!< Listening status has changed

public Q_SLOTS:
    void        loadAllSets();                                                      //!< Reload commands sets from settings
    void        listen(bool);
    QString    *recognize_from_microphone();                                        //!< Call vosk, wait for an utterance completed

protected:
    void        timerEvent(QTimerEvent *event);                                     //!< Handle mouse glide timer

protected Q_SLOTS:
    void        stopGliding();                                                      //!< Mouse glide was stopped

private:
    QList<CommandSet *>     cmdSets;
    QString                 lastCommand;
    int                     lastCount;
    QRect                   screen;
    DoToolPipe             *doTool;
    QStringList             repeaters;

    // Mouse move data
    bool                    zooming;            // Zooming mode flag
    Direction               direction;          // Mouse glide direction
    int                     timerGlide;         // Mouse is gliding when != 0

    // Vosk data
    VoskModel              *model = nullptr;
    VoskRecognizer         *recognizer = nullptr;
    bool                    listening;
    VoskStatus              status;

    // Audio device data
    QAudioFormat            format;
    QAudioSource           *audio;
    QIODevice              *microphone;
    QMediaDevices           medias;
    QAudioInput             audioInput;
};

#endif // COMPRENDO_H
