/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

#include <time.h>
#include <iostream>

#include <QApplication>
#include <QLocale>
#include <QTranslator>
#include <QCommandLineParser>
#include <QMessageBox>
#include <QLibraryInfo>
#include <QStandardPaths>
#include <QSplashScreen>

#include "app_version.h"
#include "mainwindow.h"
#include "modelchooser.h"
#include "runguard.h"
#include "misc.h"


#define EXAMPLES_DIR    "/usr/share/%1"
#define VOSK_MODEL_SYSTEM_PATH  "/usr/share/vosk-models/"
#define VOSK_MODEL_USER_DIR  "%1/vosk-models/"
#define VOSK_MODEL_URL  "https://alphacephei.com/vosk/models"
#define VOSK_MODEL_FILE "model-list.json"
#define SPEECH_LANG_DIR "/usr/share/pico/lang"
#define MOUSE_FILE      "mouseTo.nocm"
#define REPEATER_FILE   "repeater.nocr"


int main(int argc, char *argv[])
{
//    QT_REQUIRE_VERSION(argc, argv, "6.4.0")
    QApplication app(argc, argv);
    app.setOrganizationName(APP_ORG);
    app.setOrganizationDomain(APP_MAIL);
    app.setApplicationName(APP_NAME);
    app.setApplicationVersion(APP_VERSION);

#if (QT_VERSION >= QT_VERSION_CHECK(6, 0, 0))
    QString pathToLibs = QLibraryInfo::path(QLibraryInfo::TranslationsPath);
#else
    QString pathToLibs = QLibraryInfo::location(QLibraryInfo::TranslationsPath);
#endif

    // Start splash screen
    QPixmap pixmap(":/icons/nocomprendo-splash.svg");
    QSplashScreen splash(pixmap);
    splash.setWindowFlag(Qt::WindowStaysOnTopHint, true);
    splash.show();
    QCoreApplication::processEvents();      // Called several times for splashscreen

    // Load the system translations provided by Qt for message boxes, file dialogs, etc.
    QTranslator qtBaseTranslator;
    QLocale myLocale(QLocale::system().language());
    QString lang = myLocale.name();
    lang = lang.mid(0,lang.indexOf('_'));
    if (qtBaseTranslator.load(QString("qtbase_%1.qm").arg(lang), pathToLibs))
        app.installTranslator(&qtBaseTranslator);

    // Load custom translations for this application
    QTranslator translator;
    const QStringList uiLanguages = QLocale::system().uiLanguages();
    for (const QString &locale : uiLanguages) {
        const QString baseName = QString(APP_STR) + "_" + QLocale(locale).name();
        if (translator.load(":/i18n/" + baseName)) {
            app.installTranslator(&translator);
            break;
        } else {
            std::cerr << "Translation file not found for " << baseName.toStdString().c_str() << "\n";
        }
    }
    QCoreApplication::processEvents();      // Called several times for splashscreen

    // Parse command line options
    QCommandLineParser parser;
    parser.addHelpOption();
    parser.setApplicationDescription(QObject::tr("NoComprendo is a voice command and dictation tool for Linux.\nBased on Qt and Vosk."));
    parser.addVersionOption();
    QCommandLineOption authorOption(QStringList()  << "a" << "author",          QObject::tr("Show author informations."));
    parser.addOption(authorOption);
    QCommandLineOption licenseOption(QStringList() << "i" << "license",         QObject::tr("Show license informations."));
    parser.addOption(licenseOption);
    QCommandLineOption wordOption(QStringList()    << "c" << "commands",        QObject::tr("Trace recognized commands."));
    parser.addOption(wordOption);
    QCommandLineOption lmOption(QStringList()      << "l" << "languageModel",   QObject::tr("Show language model file."));
    parser.addOption(lmOption);
    parser.process(app);        // Process the actual command line arguments given by the user
    QCoreApplication::processEvents();      // Called several times for splashscreen

    // Initialize parameters
    APPCONF.lang = QLocale::system().name();                // environment variable LANGUAGE
    APPCONF.appPath         = QDir(QString(EXAMPLES_DIR).arg(APP_STR));
    APPCONF.voskSystemPath        = QDir(VOSK_MODEL_SYSTEM_PATH);
    APPCONF.voskSystemPath.setFilter(QDir::Dirs);
    APPCONF.voskUserPath = QDir(QString(VOSK_MODEL_USER_DIR).arg(QStandardPaths::writableLocation(QStandardPaths::ConfigLocation)));
    if (!APPCONF.voskUserPath.exists()) {
        APPCONF.voskUserPath.mkpath(".");
    }

    APPCONF.voskUrl = VOSK_MODEL_URL;
    APPCONF.voskJson = QString("%1/%2").arg(VOSK_MODEL_URL).arg(VOSK_MODEL_FILE);

    // Check if an other instance of nocomprendo is already running.
    RunGuard guard( "be.root-nocomprendo-bruno.anselme-2025" );
    if (!guard.tryToRun()) {
        QMessageBox::information(nullptr
                                 , QApplication::tr("%1 already running").arg(APP_NAME)
                                 , QApplication::tr("Can't start more than one instance of %1.").arg(APP_NAME)
                                 );
        exit(0);
    }
    QCoreApplication::processEvents();      // Called several times for splashscreen

    // Check for an existing default model in settings
    QSettings   settings(APP_ORG, APP_NAME);
    QString     defModel = settings.value("defModel", "").toString();
    if ((defModel == "") || (!QFile::exists(defModel))) {
        splash.hide();
        QCoreApplication::processEvents();      // Called several here to process hide()
        ModelChooser    dial;                // Language selector
        int dialogCode = dial.exec();
        if (dialogCode == QDialog::Accepted) {
            APPCONF.model = dial.getModel();
            if (APPCONF.model.isEmpty())
                exit(0);
            settings.setValue("defModel", APPCONF.model);
        } else {
            exit(0);
        }
        splash.show();
    } else {
        APPCONF.model = defModel;
    }
    QCoreApplication::processEvents();      // Called several times for splashscreen

    APPCONF.userSetsPath    = QStandardPaths::writableLocation(QStandardPaths::AppConfigLocation);
    if (!APPCONF.userSetsPath.exists())                     // Create first time
        APPCONF.userSetsPath.mkpath(APPCONF.userSetsPath.path());

    QRegularExpression re(".*vosk-model-(.*)-\\d+\\.\\d+");
    APPCONF.setsDirName = APPCONF.model;
    APPCONF.setsDirName = APPCONF.setsDirName.replace(re, "\\1");   // Contains en, fr, small-en, small-fr etc

    if (!APPCONF.userSetsPath.exists(APPCONF.setsDirName))    // Create first time
        APPCONF.userSetsPath.mkdir(APPCONF.setsDirName);
    APPCONF.userSetsPath.cd(APPCONF.setsDirName);             // Now userSetsPath should be: /home/<user>/.config/nocomprendo/en

    APPCONF.rFontSize = 18;
    APPCONF.mouseFile = QString("%1/%2").arg(APPCONF.userSetsPath.absolutePath()).arg(MOUSE_FILE);
    APPCONF.repeaterFile = QString("%1/%2").arg(APPCONF.userSetsPath.absolutePath()).arg(REPEATER_FILE);
    APPCONF.speechDir = QDir(SPEECH_LANG_DIR);
    APPCONF.speechLang = APPCONF.lang.replace("_", "-");
    APPCONF.commandTrace = parser.isSet("commands");
    QCoreApplication::processEvents();      // Called several times for splashscreen

    Speech::getSpeechesList();      // Initialize APPCONF.speechesList

//    APPCONF.trace();    exit(0);

    // Run command line options and exit
    if (parser.isSet("author")) {
        std::cout << QObject::tr("%1 was written by:\n\t%2 <%3>\n").arg(APP_STR).arg(APP_AUTHOR).arg(APP_MAIL).toStdString().c_str();
        std::cout << QObject::tr("\tWeb site : %1\n").arg(APP_SITE).toStdString().c_str();
        exit(EXIT_SUCCESS);
    }
    if (parser.isSet("license")) {
        std::cout << readFile(":/help/license-gpl-3.0.txt").toStdString().c_str();
        exit(EXIT_SUCCESS);
    }
    if (parser.isSet("languageModel")) {
        std::cout << "Language model:" << defModel.toStdString().c_str() << "\n";
        exit(EXIT_SUCCESS);
    }

    // Check for write access on uinput was here (dotool: https://git.sr.ht/~geb/dotool/tree/HEAD/doc/dotool.1.scd).
    // Replaced by udev rule. Need reboot to be applied (after dotool install).

    MainWindow w;
    w.show();

    // 5 seconds timeout for splash-screen test
    // QTime dieTime= QTime::currentTime().addSecs(5);
    // while (QTime::currentTime() < dieTime)
    //     QCoreApplication::processEvents(QEventLoop::AllEvents, 100);

    splash.finish(&w);
    return app.exec();
}
