/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

/**
 * @file      misc.cpp
 * @author    Bruno ANSELME
 * @brief     Miscellaneous functions
 * @details   Read and write file. get desktop rectangle.
 */

#include <QFile>
#include <QTextStream>
#include <QGuiApplication>
#include <QScreen>
#include <QProcess>
#include "misc.h"

// Receive a name, returns a to an executable file. Search in /usr/bin or ask with which.
QString findProgram(QString name) {
    QString prog("/usr/bin/" + name);
    if (!QFile::exists(prog)) {
        QProcess searchProgramProcess;
        searchProgramProcess.start("which", { name });
        searchProgramProcess.waitForFinished();
        if (searchProgramProcess.exitCode()) {
            qWarning() << name << "is not avalaible";
            exit(EXIT_FAILURE);
        } else {
            prog = searchProgramProcess.readAllStandardOutput().trimmed();
            qWarning() << name << "found here:" << prog;
        }
    }
    return prog;
}

QString readFile(QString fname) {
    QFile file(fname);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
        return QString();
    QTextStream f(&file);
    QString txt = f.readAll();
    file.close();
    return txt;
}

void writeFile(QString fname, QString &txt) {
    QFile file(fname);
    if (!file.open(QIODevice::WriteOnly | QIODevice::Text))
        return;
    QTextStream f(&file);
    f << txt;
    file.close();
}

void writeFile(QString fname, QByteArray &bytes) {
    QFile file(fname);
    if (!file.open(QIODevice::WriteOnly))
        return;
    file.write(bytes);
    file.close();
}

QRect getDesktopRect() {
    int   xmax = 0;
    int   ymax = 0;
    QList<QScreen *> screens = QGuiApplication::screens();
    for (int i = 0; i < screens.count(); i++) {
        QRect r = screens[i]->geometry();
        xmax = qMax(xmax, r.right());
        ymax = qMax(ymax, r.bottom());
    }
    return QRect(0, 0, xmax, ymax);
}

QStringList parseCombinedArgString(const QString &program)
{
    QStringList args;
    QString tmp;
    int quoteCount = 0;
    bool inQuote = false;
    // handle quoting. tokens can be surrounded by double quotes
    // "hello world". three consecutive double quotes represent
    // the quote character itself.
    for (int i = 0; i < program.size(); ++i)
    {
        if (program.at(i) == QLatin1Char('"'))
        {
            ++quoteCount;
            if (quoteCount == 3)
            {
                // third consecutive quote
                quoteCount = 0;
                tmp += program.at(i);
            }
            continue;
        }
        if (quoteCount)
        {
            if (quoteCount == 1)
                inQuote = !inQuote;
            quoteCount = 0;
        }
        if (!inQuote && program.at(i).isSpace())
        {
            if (!tmp.isEmpty())
            {
                args += tmp;
                tmp.clear();
            }
        }
        else
        {
            tmp += program.at(i);
        }
    }
    if (!tmp.isEmpty())
        args += tmp;
    return args;
}
