/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

#include <QJsonDocument>
#include <QJsonObject>
#include <QProcess>
#include <QPushButton>
#include <QStorageInfo>
#include <QMessageBox>

#include "app_version.h"
#include "modelloader.h"
#include "misc.h"
#include "ui_modelloader.h"

ModelLoader::ModelLoader() :
    QDialog(),
    ui(new Ui::dialogModelLoader)
  , downloader(nullptr)
    , systemSpace(false)
{
    ui->setupUi(this);
    setWindowModality(Qt::ApplicationModal);
    ui->lbTitle->setText(APPCONF.voskUrl.toString());
    jsonName = QString("%1/model-list.json").arg(APPCONF.voskUserPath.path());
    downloader = new FileDownloader(APPCONF.voskJson, jsonName, this);
    connect(downloader, SIGNAL(downloaded()), this, SLOT(fillLanguageView()));
    connect(ui->buttonBox, SIGNAL(accepted()), this, SLOT(close()));
    connect(ui->btClose, SIGNAL(clicked()), this, SLOT(close()));
    connect(ui->btSystemSpace, SIGNAL(clicked()), this, SLOT(loadSystemModel()));
    connect(ui->btUserSpace, SIGNAL(clicked()), this, SLOT(loadUserModel()));
}

ModelLoader::~ModelLoader()
{
    delete ui;
}

void ModelLoader::loadModel() {
    if (ui->twLanguages->currentRow() > -1)
        loadModel(ui->twLanguages->currentRow(), 0);
}

void ModelLoader::loadModel(int row, int) {
    QTableWidgetItem *item = ui->twLanguages->item(row, 1);
    urlModel = QString("%1/%2.zip").arg(APPCONF.voskUrl.toString()).arg(item->text());
    fileName = QString("%1/%2.zip").arg(APPCONF.voskUserPath.path()).arg(item->text());
    if (QFile::exists(fileName)) {
        // qWarning() << fileName << "already loaded";
        unzipModel();
    } else {
        // Check for sufficient disk space
        QStorageInfo info(APPCONF.voskUserPath);
        QString strSize = ui->twLanguages->item(row, 3)->text();
        qint64 zipSize = 0;
        QRegularExpression re("(\\d+\\.\\d)([MG])iB");
        QRegularExpressionMatch match = re.match(strSize);
        if (match.hasMatch()) {
            float floatSize = match.captured(1).toFloat();
            qint64 ratio = (match.captured(2) == "M") ? 1024 * 1024 : (match.captured(2) == "G") ? 1024 * 1024 * 1024 : 0;
            zipSize = floatSize * ratio;
        }
        qint64 requiredBytes = zipSize * 3;     // Required bytes on disk, once for zip file and twice for future expanded file
        if (info.bytesFree() < requiredBytes) {
            QMessageBox::information(nullptr
                                     , QApplication::tr("%1 error").arg(APP_NAME)
                                     , tr("Insufficient disk space to download and unzip the language model %1").arg(item->text())
                                     );
            return;
        }
        // Start download
        ui->lbCurrentAction->setText(tr("Loading model"));
        ui->lbCurrentAction->repaint();
        downloader = new FileDownloader(urlModel, fileName, this);
        connect(downloader, SIGNAL(downloaded()), this, SLOT(unzipModel()));
        connect(downloader, SIGNAL(aborted()), this, SLOT(abortDownload()));
        // qWarning() << "Loading" << urlModel;
    }
}

void ModelLoader::abortDownload() {
}

void ModelLoader::unzipModel() {
    // Unzip downloaded language model
    ui->btSystemSpace->setEnabled(false);
    ui->btUserSpace->setEnabled(false);
    ui->btClose->setEnabled(false);
    QString unzip = findProgram("unzip");
    ui->lbCurrentAction->setText(tr("Extracting model, please wait..."));
    ui->lbCurrentAction->repaint();
    QProcess aProcess;
    if (systemSpace) {
        QString pkexec = findProgram("pkexec");
        aProcess.setProgram(pkexec);
        aProcess.setArguments({ unzip, fileName, "-d", APPCONF.voskSystemPath.path() });
    } else {
        aProcess.setProgram(unzip);
        aProcess.setArguments({ fileName, "-d", APPCONF.voskUserPath.path() });
    }
    aProcess.start();
    aProcess.waitForFinished(-1);
    ui->lbCurrentAction->setText(tr("Model ready"));
    ui->lbCurrentAction->repaint();
    QFile::remove(fileName);
    // close();
    ui->btSystemSpace->setEnabled(true);
    ui->btUserSpace->setEnabled(true);
    ui->btClose->setEnabled(true);
}

void ModelLoader::fillLanguageView() {
    disconnect(downloader, SIGNAL (downloaded()), this, SLOT (fillLanguageView()));
    QString answer = readFile(jsonName);
    QFile::remove(jsonName);

    if (downloader->error()) {
        ui->teError->setText(QString("<h4>%1</h4>").arg(downloader->errorString()));
        ui->stack->setCurrentIndex(1);
        downloader->deleteLater();
        return;
    }

    QJsonDocument doc(QJsonDocument::fromJson(answer.toUtf8()));                   // Load json document
    jsonModels    = doc.array();
    ui->twLanguages->verticalHeader()->hide();
    int count = 0;
    for (int i = 0; i < jsonModels.count(); i++) {
        ui->twLanguages->setRowCount(count + 1);
        // Remove some models
        QString type = jsonModels[i].toObject()["type"].toString();
        if (type == "tts") continue;            // Unknow type
        QString obsolete = jsonModels[i].toObject()["obsolete"].toString();
        if (obsolete == "true") continue;       // Obsolete model
        Qt::ItemFlags flags = Qt::ItemIsSelectable | Qt::ItemIsEnabled;

        QDir dirUser(APPCONF.voskUserPath.path() + "/" + jsonModels[i].toObject()["name"].toString());
        if (dirUser.exists()) flags = Qt::NoItemFlags;;             // Already loaded model. Shown but not selectable
        QDir dirSystem(APPCONF.voskSystemPath.path() + "/" + jsonModels[i].toObject()["name"].toString());
        if (dirSystem.exists()) flags = Qt::NoItemFlags;;             // Already loaded model. Shown but not selectable

        // Add model informations to table
        QTableWidgetItem *newItem;
        newItem = new QTableWidgetItem(jsonModels[i].toObject()["lang_text"].toString());
        newItem->setFlags(flags);
        ui->twLanguages->setItem(count, 0, newItem);
        newItem = new QTableWidgetItem(jsonModels[i].toObject()["name"].toString());
        newItem->setFlags(flags);
        ui->twLanguages->setItem(count, 1, newItem);
        newItem = new QTableWidgetItem(jsonModels[i].toObject()["version"].toString());
        newItem->setFlags(flags);
        ui->twLanguages->setItem(count, 2, newItem);
        newItem = new QTableWidgetItem(jsonModels[i].toObject()["size_text"].toString());
        newItem->setTextAlignment(Qt::AlignRight | Qt::AlignVCenter);
        newItem->setFlags(flags);
        ui->twLanguages->setItem(count, 3, newItem);
        newItem = new QTableWidgetItem(type);
        newItem->setFlags(flags);
        ui->twLanguages->setItem(count, 4, newItem);
        count++;
    }
    ui->twLanguages->sortByColumn(0, Qt::AscendingOrder);
    ui->twLanguages->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    ui->twLanguages->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    downloader->deleteLater();
}

void ModelLoader::changeEvent(QEvent *e)
{
    QDialog::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}
