/*
 * SPDX-FileCopyrightText: 2024-2025 Bruno Anselme <be.root@gfree.fr>
 *
 * Authors:
 *   Bruno Anselme <be.root@gfree.fr>
 *
 *   SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of the NoComprendo software.
*/

/**
 * @file      trayicon.cpp
 * @author    Bruno ANSELME
 * @brief     TrayIcon class definition
 * @details   System tray icon for NoComprendo
 */

#include "app_version.h"
#include "trayicon.h"

TrayIcon::TrayIcon() : QSystemTrayIcon()
    , trayStatus(Stopped)
{
    setTheme("color-dark");
    messages[Stopped]       = tr("%1 stopped").arg(APP_NAME);
    messages[Ready]         = tr("%1 ready").arg(APP_NAME);
    messages[Listening]     = tr("%1 listening").arg(APP_NAME);
    messages[Suspended]     = tr("%1 suspended").arg(APP_NAME);
    messages[ListenSuspend] = tr("%1 listening").arg(APP_NAME);

    setIcon(icons[Stopped]);

    QAction    *quitAction, *showAction;
    showAction  = myMenu.addAction(tr("Show"));

    actions[Ready]      = myMenu.addAction(QIcon(":/icons/start.png"), tr("Start"));
    actions[Stopped]    = myMenu.addAction(QIcon(":/icons/stop.png"),  tr("Stop"));
    actions[Suspended]  = myMenu.addAction(QIcon(":/icons/suspend.png"), tr("Suspend"));
    myMenu.addSeparator();

    actions[Dictation]  = myMenu.addAction(QIcon(":/icons/nocomprendo.png"), tr("Dictation"));
    actions[Dictation]->setCheckable(true);
    actions[Speak]  = myMenu.addAction(QIcon(":/icons/speaker.png"), tr("Clipboard to speech"));
    myMenu.addSeparator();

    quitAction  = myMenu.addAction(QIcon(":/icons/exit.png"), tr("Quit %1").arg(APP_NAME));

    connect(actions[Dictation], SIGNAL(triggered()), this,  SLOT(sendDictation()));
    connect(actions[Speak],     SIGNAL(triggered()), this,  SLOT(sendSpeak()));
    connect(actions[Ready],     SIGNAL(triggered()), this,  SLOT(sendStart()));
    connect(actions[Stopped],   SIGNAL(triggered()), this,  SLOT(sendStop()));
    connect(actions[Suspended], SIGNAL(triggered()), this,  SLOT(sendSuspend()));
    connect(quitAction,         SIGNAL(triggered()), this,  SLOT(sendQuit()));
    connect(showAction,         SIGNAL(triggered()), this,  SLOT(sendShow()));
    setContextMenu(&myMenu);
    show();
}

void TrayIcon::setTheme(QString themeDir_) {
    themeDir = ":/themes/" + themeDir_;
    icons[Stopped]          = QIcon(themeDir + "/stop.svg");
    icons[Ready]            = QIcon(themeDir + "/ready.svg");
    icons[Listening]        = QIcon(themeDir + "/listen-ready.svg");
    icons[Suspended]        = QIcon(themeDir + "/suspend.svg");
    icons[ListenSuspend]    = QIcon(themeDir + "/listen-suspend.svg");
    icons[Dictation]        = QIcon(themeDir + "/dictate.svg");
    icons[ListenDictation]  = QIcon(themeDir + "/listen-dictate.svg");
}

void TrayIcon::setDictationChecked(bool checked) {
    actions[Dictation]->setChecked(checked);
}

void TrayIcon::setMenus(bool running, bool suspended) {
    actions[Ready]->setEnabled(!running);
    actions[Stopped]->setEnabled(running);
    actions[Dictation]->setEnabled(running && !suspended);
    actions[Suspended]->setEnabled(running);
    actions[Suspended]->setText((running && !suspended) ? tr("Suspend") : tr("Resume") );
}

void TrayIcon::setStatus(Status s) {
    if (trayStatus != s) {
        trayStatus = s;
        setIcon(icons[trayStatus]);
        setToolTip(messages[trayStatus]);
    }
}

void TrayIcon::sendDictation()  { emit trayDictation(); }
void TrayIcon::sendSpeak()      { emit traySpeak();     }
void TrayIcon::sendStart()      { emit trayStart();     }
void TrayIcon::sendStop()       { emit trayStop();      }
void TrayIcon::sendSuspend()    { emit traySuspend();   }
void TrayIcon::sendQuit()       { emit trayQuit();      }
void TrayIcon::sendShow()       { emit trayShow();      }

